"""Helper functions to compute hierarchical evaluation metrics."""
import numpy as np


def precision(y_true: np.ndarray, y_pred: np.ndarray):
    """
    Compute precision score for hierarchical classification.

    hP = sum(|S intersection T|) / sum(|S|),
    where S is the set consisting of the most specific class(es) predicted for a test example and all respective ancestors
    and T is the set consisting of the true most specific class(es) for a test example and all respective ancestors.

    Parameters
    ----------
    y_true : np.array of shape (n_samples, n_levels)
        Ground truth (correct) labels.
    y_pred : np.array of shape (n_samples, n_levels)
        Predicted labels, as returned by a classifier.
    Returns
    -------
    precision : float
        What proportion of positive identifications was actually correct?
    """
    assert len(y_true) == len(y_pred)
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        sum_intersection = sum_intersection + len(
            set(ground_truth).intersection(set(prediction))
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(
            set(prediction)
        )
    precision = sum_intersection / sum_prediction_and_ancestors
    return precision


def recall(y_true: np.ndarray, y_pred: np.ndarray):
    """
    Compute recall score for hierarchical classification.

    hR = sum(|S intersection T|) / sum(|T|),
    where S is the set consisting of the most specific class(es) predicted for a test example and all respective ancestors
    and T is the set consisting of the true most specific class(es) for a test example and all respective ancestors.

    Parameters
    ----------
    y_true : np.array of shape (n_samples, n_levels)
        Ground truth (correct) labels.
    y_pred : np.array of shape (n_samples, n_levels)
        Predicted labels, as returned by a classifier.
    Returns
    -------
    recall : float
        What proportion of actual positives was identified correctly?
    """
    assert len(y_true) == len(y_pred)
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        sum_intersection = sum_intersection + len(
            set(ground_truth).intersection(set(prediction))
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(
            set(ground_truth)
        )
    recall = sum_intersection / sum_prediction_and_ancestors
    return recall


def f1(y_true: np.ndarray, y_pred: np.ndarray):
    """
    Compute f1 score for hierarchical classification.

    hF = 2 * hP * hR / (hP + hR),
    where hP is the hierarchical precision and hR is the hierarchical recall.

    Parameters
    ----------
    y_true : np.array of shape (n_samples, n_levels)
        Ground truth (correct) labels.
    y_pred : np.array of shape (n_samples, n_levels)
        Predicted labels, as returned by a classifier.
    Returns
    -------
    f1 : float
        Weighted average of the precision and recall
    """
    assert len(y_true) == len(y_pred)
    prec = precision(y_true, y_pred)
    rec = recall(y_true, y_pred)
    f1 = 2 * prec * rec / (prec + rec)
    return f1



"""
Local classifier per parent approach.

Numeric and string output labels are both handled.
"""

from itertools import repeat
from typing import Any, Tuple, Optional, Union

import networkx as nx
import numpy as np
from sklearn.base import BaseEstimator

from hiclass import BinaryPolicies
from hiclass.BinaryPolicies import Policy
from hiclass.Classifier import (
    DuplicateFilter,
    NodeClassifier,
    ConstantClassifier,
    fit_or_replace,
)
from hiclass.data import NODE_TYPE


def _fit_classifier(
    node: NODE_TYPE,
    classifier: Any,
    hierarchy: nx.DiGraph,
    policy: Policy,
    data: np.ndarray,
    replace_classifier: bool,
) -> Tuple[NODE_TYPE, Any, str]:
    """Fits a local classifier for a given node.

    Parameters
    ----------
    node : int or str
        Node for which the classifier was fitted.
    classifier : Any
        The classifier that is assigned to the given node.
    policy : Policy
        Rules for defining positive and negative training samples.
    data : {array-like, sparse matrix} of shape (n_samples, n_features)
        The training input samples. If a sparse matrix is provided, it will be
        converted into a sparse ``csc_matrix``.
    replace_classifier : bool
        Turns on (True) the replacement of classifiers with a constant classifier when trained on only
        a single unique class.

    Returns
    -------
    node : int or str
        Node for which the classifier was fitted.
    classifier : Any
        The classifier that was fitted for the given node.
    warning : str
        The warnings raised.
    """
    classes = np.zeros_like(policy.data, dtype=int)
    all_samples = np.zeros_like(policy.data, dtype=bool)

    successors = list(hierarchy.successors(node))
    mapping = dict(zip(successors, range(0, len(successors))))
    for successor in successors:
        positive_samples = policy.positive_samples(successor)
        classes[positive_samples] = mapping[successor]
        all_samples[positive_samples] = True

    classifier, warning = fit_or_replace(
        replace_classifier,
        classifier,
        data[all_samples],
        classes[all_samples],
        len(mapping),
    )
    return node, classifier, warning


class LocalClassifierPerParent(NodeClassifier):
    """
    Assign local classifiers to each parent node in a DAG.

    A local classifier per parent is a local hierarchical classifier that fits one local multi-class classifier
    for each node that has children. The labels correspond to its children.
    """

    def __init__(
        self,
        n_jobs: int = 1,
        verbose: int = 0,
        local_classifier: Any = BaseEstimator(),
        hierarchy: Optional[nx.DiGraph] = None,
        unique_taxonomy: bool = True
    ) -> None:
        """
        Initialize a classifier and set the :code:`policy` parameter of its superclass to :code:`SiblingsPolicy`.

        Parameters
        ----------
        n_jobs : int, default=1
            The number of jobs to run in parallel. Only `fit` is parallelized.
        verbose : int, default=0
            Controls the verbosity when fitting and predicting.
            See https://verboselogs.readthedocs.io/en/latest/readme.html#overview-of-logging-levels
            for more information.
        local_classifier : BaseEstimator instance
            The local classifier used to create the collection of local classifier. Needs to have fit, predict and clone
            methods.
        hierarchy : nx.DiGraph, default=None
            Label hierarchy used in prediction and fitting. If None, it will be inferred during training.
        unique_taxonomy : bool, default=True
            True if the elements in the hierarchy have unique names, otherwise it can have unexpected behaviour.
            For example, a->b->c and d->b->e could have different meanings for b, so in that case unique_taxonomy
            should be set to false.
        """
        super().__init__(
            n_jobs=n_jobs,
            verbose=verbose,
            local_classifier=local_classifier,
            hierarchy=hierarchy,
            unique_taxonomy=unique_taxonomy,
            policy=BinaryPolicies.SiblingsPolicy,
        )

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        placeholder_label=None,
        replace_classifiers=True,
    ) -> None:
        """
        Fit the local classifiers.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. If a sparse matrix is provided, it will be
            converted into a sparse ``csc_matrix``.
        Y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The multi-class hierarchical labels.
        placeholder_label : int or str, default=None
            Label that corresponds to "no label available for this data point".
        replace_classifiers : bool, default=True
            Turns on (True) the replacement of classifiers with a constant classifier when trained on only
            a single unique class.
        """
        super().fit(X, Y, placeholder_label, replace_classifiers)

        duplicate_filter = DuplicateFilter()
        self.log.addFilter(duplicate_filter)  # removes possible duplicate warnings

        nodes = list(nx.dfs_preorder_nodes(self.hierarchy, self.root))
        leaf_nodes = [
            x for x in self.hierarchy.nodes() if self.hierarchy.out_degree(x) == 0
        ]
        for leaf in leaf_nodes:
            nodes.remove(leaf)
        classifiers = [self.get_classifier(node) for node in nodes]
        fit_args = zip(
            nodes,
            classifiers,
            repeat(self.hierarchy),
            repeat(self.policy),
            repeat(X),
            repeat(replace_classifiers),
        )
        self._fit_classifier(_fit_classifier, fit_args)

        self.log.removeFilter(duplicate_filter)  # delete duplicate filter

    def _create_prediction(
        self, data: np.ndarray, selected_rows: np.ndarray, parent_label: NODE_TYPE
    ) -> np.ndarray:
        classifier = self.get_classifier(parent_label)
        return self._predict_with_classifier(classifier, data[selected_rows]).T

    def _get_nodes_with_classifier(self):
        return [x for x in self.hierarchy.nodes() if self.hierarchy.out_degree(x) != 0]



"""
Local classifier per node approach.

Numeric and string output labels are both handled.
"""
from itertools import repeat
from typing import Any, List, Tuple

import networkx as nx
import numpy as np

from hiclass.BinaryPolicies import Policy
from hiclass.Classifier import (
    DuplicateFilter,
    NodeClassifier,
    fit_or_replace,
)
from hiclass.data import NODE_TYPE


def _fit_classifier(
    node: NODE_TYPE,
    classifier: Any,
    policy: Policy,
    data: np.ndarray,
    replace_classifier: bool,
) -> Tuple[NODE_TYPE, Any, str]:
    """Fits a local classifier for a given node, policy and data.

    Parameters
    ----------
    node : int or str
        Node for which the classifier was fitted.
    classifier : Any
        The classifier that is assigned to the given node.
    policy : Policy
        Rules for defining positive and negative training samples.
    data : np.array
        The data that is being used for training the classifier.
    replace_classifier : bool
        Turns on (True) the replacement of classifiers with a constant classifier when trained on only
        a single unique label.

    Returns
    -------
    node : int or str
        Node for which the classifier was fitted.
    classifier : Any
        The classifier that was fitted for the given node.
    warning : str
        The warnings raised.
    """
    classes = np.zeros_like(policy.data, dtype=bool)
    positive_samples = policy.positive_samples(node)
    negative_samples = policy.negative_samples(node)
    all_samples = np.logical_or(positive_samples, negative_samples)
    classes[positive_samples] = 1

    classifier, warning = fit_or_replace(
        replace_classifier, classifier, data[all_samples], classes[all_samples], 2
    )
    return node, classifier, warning


class LocalClassifierPerNode(NodeClassifier):
    """
    Assign local classifiers to each node of the graph, except the root node.

    A local classifier per node is a local hierarchical classifier that fits one local binary classifier
    for each node of the class hierarchy, except for the root node.
    """

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        placeholder_label=None,
        replace_classifiers=True,
    ) -> None:
        """
        Fit the local classifiers.

        Parameters
        ----------
        X : np.array of shape(n_samples, n_features)
            The training input samples.
        Y : np.array of shape (n_samples, n_levels)
            The hierarchical labels.
        placeholder_label : int or str, default=None
            Label that corresponds to "no label available for this data point". Defaults will be used if not passed.
        replace_classifiers : bool, default=True
            Turns on (True) the replacement of a local classifier with a constant classifier when trained on only
            a single unique class.
        """
        super().fit(X, Y, placeholder_label, replace_classifiers)

        duplicate_filter = DuplicateFilter()
        self.log.addFilter(duplicate_filter)  # removes possible duplicate warnings

        nodes = list(nx.dfs_preorder_nodes(self.hierarchy, self.root))
        nodes.remove(self.root)
        classifiers = [self.get_classifier(node) for node in nodes]
        fit_args = zip(
            nodes,
            classifiers,
            repeat(self.policy),
            repeat(X),
            repeat(replace_classifiers),
        )

        self._fit_classifier(_fit_classifier, fit_args)

        self.log.removeFilter(duplicate_filter)  # delete duplicate filter

    def _choose_required_prediction_args(
        self,
        local_identifier: NODE_TYPE,
        nodes_to_predict: List[NODE_TYPE],
        selected_rows: np.ndarray,
    ) -> Tuple[np.ndarray, List[NODE_TYPE]]:
        return selected_rows, nodes_to_predict

    def _create_prediction(
        self, data: np.ndarray, selected_rows: np.ndarray, nodes_to_predict: np.ndarray
    ) -> np.ndarray:
        predictions = np.empty(
            (len(nodes_to_predict), np.sum(selected_rows)), dtype=np.float32
        )
        for i in range(len(nodes_to_predict)):
            # take only the probability of true class
            classifier = self.get_classifier(nodes_to_predict[i])
            predictions[i] = self._predict_with_classifier(
                classifier, data[selected_rows]
            )[:, 1]
        return predictions

    def _get_nodes_with_classifier(self) -> List[NODE_TYPE]:
        return [x for x in self.hierarchy.nodes() if self.hierarchy.in_degree(x) != 0]
